
/*
  This  code insert handles histogramming functions
   This assumes that the data is in TDR format
   Also that the USEINT option is used  (see execV*.c)
*/

#include <stdint.h>
#include <math.h>

/* --------------------------------------------------------------------- */

void FEBEXhistogram_data ()
{
    int len;
    int group, item;
    unsigned int chan;
    unsigned int * eventbuffer;
    unsigned int data, header, timestamp;
    int count;
    int i;
    
    int module;
    int link;
    int adc;
    int channel;
    
    unsigned short icode;
    unsigned int ifield;
    
    uint16_t signbit, exponent, mantissa;
    
    int16_t f16;
    
    float dataf;

    
    int GroupBase;
    
    
    if (*dacq_Trace > 3) {
    
    len = EventBufferLen;           //   len in units of int (multiple of 4)
    eventbuffer = (unsigned int *) EventBufferAddr;

    printf("histogram event: length %d\n",len);
    if (len == 0) {return;}  
    if (len > 32) len=32;
    while (len > 0) {printf("0x%lx ",*eventbuffer++); len--;}
    printf("\n\n");
    
    }


    if (*dacq_HistEnable == 0) {return;}   /*  histogramming not enabled  */

#ifdef USETHREADS
    len = EventBufferLen[ID];             /*   len in units of int  */
    eventbuffer = (unsigned int *) &EventBuffer[ID][0];
#else
    len = EventBufferLen;             /*   len in units of int  */
    eventbuffer = (unsigned int *) EventBufferAddr;
#endif

    if (len == 0) {return;}     /* empty event */

//    if(*dacq_Trace > 2) printf("H1: len=%d\n",len);

    while (len > 1)
    {

        data = *eventbuffer++;
	timestamp = *eventbuffer++;
	
	len -=2;

//        if(*dacq_Trace > 2)printf("H2: len=%d; data=0x%x 0x%x\n",len,data,timestamp);

        switch ((data >> 30 ) & 3) {

	   case  1:    /* wave form */ 
	     
		module = (data >> 22) & 15;   /*  board id */
		link = (data >> 26) & 3;      /*  sfp id  */
		adc = (data >> 16) & 15;      /*  chan id  */

		group = (link * 256) + (module * 16) + adc + 1 + 512;

                count = (data + 3) & 0x3ffc;   /* round up to 4 */

/* digital trace */

if (*dacq_Trace > 2) printf("hist wave buffer group=%d %d %d %d %x %d\n", group, module, adc, count, hist_table[group][item].base, hist_table[group][item].size);
                if (hist_table[0][0].base != NULL) {
		        chan = (group << 6) + item - hist_table[0][0].offset;
			if (hist_table[0][0].shift > 0) chan = (chan >> hist_table[0][0].shift);
			if (chan < hist_table[0][0].size) hist_table[0][0].base[chan]++;
		}
		
		if (FEBEX_MWDFormat == 1) {    /* trace using data in MWD format  */

		for (i = 0; i < count; i += 2)
		{
			data = *eventbuffer++;
			len--;
			if (hist_table[group][item].base != NULL && i < hist_table[group][item].size)
			{
			  
			        f16 = (unsigned short) ((data>>16) & 0x0000ffff);
				
		                signbit = (f16 >> 15) & 0x1;
                                exponent = (f16 >> 10) & 0x1f;
                                mantissa = (f16 & 0x3ff);
				
				dataf = pow(-1,signbit) * pow(2,exponent) * (1.0 + ((float)mantissa/1024.0));
			  
				hist_table[group][item].base[i] = roundf(dataf);
				
			        f16 = (unsigned short) (data & 0x0000ffff);
				
		                signbit = (f16 >> 15) & 0x1;
                                exponent = (f16 >> 10) & 0x1f;
                                mantissa = (f16 & 0x3ff);
				
				dataf = pow(-1,signbit) * pow(2,exponent) * (1.0 + ((float)mantissa/1024.0));
    
                                hist_table[group][item].base[i+1] = roundf(dataf);
			}	  
		}
		}
		
		if (FEBEX_MWDFormat == 0)  {    /*  trace using data in orginal ADC mode   */
		
		for (i = 0; i < count; i += 2)
		{
			data = *eventbuffer++;
			len--;
			if (hist_table[group][item].base != NULL && i < hist_table[group][item].size)
			{
				hist_table[group][item].base[i] = (unsigned short) ((data>>16) & 0x0000ffff);
				hist_table[group][item].base[i+1] = (unsigned short) (data & 0x0000ffff);
			}
		}
		}
		
                if (hist_table[group][item].base != NULL) {Stats[INC]++;}


		break;
				
				
            case 3:             /*  Energy & Time  */
	      
		module = (data >> 22) & 15;   /*  board id */
		link = (data >> 26) & 3;      /*  sfp id  */
		adc = (data >> 16) & 15;      /*  chan id  */

		group = (link * 256) + (module * 16) + adc + 1;
		item = (data >> 20) & 3;
		
	        if (hist_table[0][0].base != NULL) {
	            chan = (group << 6) + item - hist_table[0][0].offset;
	            if (hist_table[0][0].shift > 0) chan = (chan >> hist_table[0][0].shift);
	            if (chan < hist_table[0][0].size) {hist_table[0][0].base[chan]++;}
		}

//	if (*dacq_Trace > 2) printf("H5.1: 0x%x 0x%x 0x%x 0x%x 0x%x\n",data, module, adc, group, item);
		
	        if (group > MAX_GROUP || item > MAX_ITEM) {continue;}
	        if (group < MIN_GROUP || item < MIN_ITEM) {continue;}

	        if (hist_table[group][item].base != NULL) {
				
	            chan = (data & 0x0000ffff) - hist_table[group][item].offset;
	            if (hist_table[group][item].shift > 0) chan = (chan >> hist_table[group][item].shift);
	            if (chan >= hist_table[group][item].size) {
                	if (*dacq_Trace > 2) printf("H6.1: 0x%x 0x%x 0x%x 0x%x\n", group, item, hist_table[group][item].size, chan);
			chan=hist_table[group][item].size - 1; Stats[BADDATA]++;
		    } else {
	                hist_table[group][item].base[chan]++; Stats[INC]++; 
//	if (*dacq_Trace > 2) printf("H6.2: 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n", data, module, adc, group, item, chan);
		    }
	        }
                
                break;
	
             case 2:    /*  information data word */

                icode = (data >> 20) & 0x0f;    /*  information code */
                ifield = data & 0x000fffff;     /*  information field */
                
		module = (data >> 24) & 15;   /*  board id */
		link = (data >> 28) & 3;      /*  sfp id  */

                switch (icode) {

		  case 7:
		  case 8:
                       break;

                    default:        /* undefined code */
                       len = 0;
                       if (*dacq_Trace > 3) printf("unexpected information data word while histogramming (0x%08x 0x%08x)\n", data, timestamp);
/*
                       *dacq_HistEnable = *dacq_HistEnable + 1;
                       if (*dacq_HistEnable > MAXHISTERRORS) {
                           printf("too many errors occuring - histogramming disabled\n");
                           *dacq_HistEnable = 0;
                       }
*/
                       break;
                }
                
                break;
        

	    case 0:   /*  bad format */
             default:
        if(*dacq_Trace > 2)printf("H3: bad event data: len=%d; data=0x%x 0x%x\n",len,data,timestamp);
	        
                len = 0;    /*   abort this buffer  */

                break;
        }
    }
   
}


/* --------------------------------------------------------------------- */
