/*
 * Vic Pucknell & Carl Unsworth STFC Daresbury Aug 2020
 * 
 * with acknowledgement 
 * J.Adamczewski-Musch, gsi, 12-May_2014
 *
 */

//   To build program for FEBEX data readout
//    gcc FEBEXexecV1.c -lc /MIDAS/lib_Linux64/libmbspex.so


#include <stdio.h>
#include <errno.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdlib.h>


#include <string.h>
#include <stdarg.h>

#include "pexor_gosip.h"
#include "libmbspex.h"

/* logical numer of pexor device*/
#define PEXDEVNO 0

#define PEX_MEM_OFF       0x100000
#define PCI_BAR0_SIZE     0x800000      // in bytes

#define PEX_SFP_OFF       0x40000

  
  static int nsfp = NUM_SFP;
  static int nslaves[NUM_SFP] = {0,0,0,0};
  static int bufid[NUM_SFP] = {0,0,0,0};

   static int sfp_channel = 0;
   
   static int nslave = NUM_ADCS;
   static int slaveID = 0;

   static int l_status;
   static int tok_mode = 0;    /*  set to 2 for running with wait for data  */

   static long sfp_phys_mem_base[NUM_SFP];

   static size_t len;    /*  units 32 bits  */
   static int int16;     /*  units 16 bits  */
   static off_t off;
   static unsigned long status;
    
    char *pa, *pe;

    int *pdat;
    
    int EOD = 0x5a5a5a5a;
    unsigned int eod;
    
    int febexfifobufferlength;
    
    

void febex_term_connection ()
{
  
    if (FEBEX_PexorEnable == 1) {
    
       if (fd_pex != 0) {close(fd_pex); fd_pex = 0;}
    }
    
    if (FEBEX_Data_In_Enable == 1) {
      
      if (FEBEX_DataInHandle != 0) {fclose (FEBEX_DataInHandle); FEBEX_DataInHandle = 0;}
       
    }
    
    if (FEBEX_Data_Out_Enable == 1) {
      
       if (FEBEX_DataOutHandle != 0) {fclose (FEBEX_DataOutHandle); FEBEX_DataOutHandle = 0;}

    }
 
}

void febex_go_connection ()
{
     int rc;
  
    if (FEBEX_Data_Out_Enable == 1) {
      
      if (FEBEX_DataOutHandle != 0) {fclose (FEBEX_DataOutHandle); FEBEX_DataOutHandle = 0;}
      
      sprintf(fname,"%s/%s",FEBEX_Data_Base,FEBEX_DataOut_File);
      FEBEX_DataOutHandle = fopen (fname,"w");
      
      if (FEBEX_DataOutHandle <= 0) {
	   printf("ERROR>> open file %s (write) \n", fname);
        } else {
	   printf ("Out File = %s\n", fname);
      }
      
    }  
  
}

void febex_init_connection () 
{
    int rc;
    int module = 0;
    
/*    obtain option values from their NetVars  */
    
    readNetInt("FEBEX_PexorEnable", 0, 1, &FEBEX_PexorEnable);
    readNetInt("FEBEX_Data_In_Enable", 0, 1, &FEBEX_Data_In_Enable);
    readNetInt("FEBEX_Data_Out_Enable", 0, 1, &FEBEX_Data_Out_Enable);
    
    readNetInt("FEBEX_DataFormat", 0, 1, &FEBEX_DataFormat);
    readNetInt("FEBEX_Do_DataRead", 0, 1, &FEBEX_Do_DataRead);
    readNetInt("FEBEX_Do_DataFormat", 0, 1, &FEBEX_Do_DataFormat);
    
    readNetInt("FEBEX_TestPatternMode", 0, 1, &FEBEX_TestPatternMode);
    readNetInt("FEBEX_TokenMode", 0, 1, &FEBEX_TokenMode);   tok_mode = FEBEX_TokenMode;
    
    readNetInt("FEBEX_MWDFormat", 0, 1, &FEBEX_MWDFormat);
    readNetInt("FEBEX_ReadOutDelay", 0, 1, &FEBEX_ReadOutDelay);

    readNetInt("FEBEX_EnergyLow", 0, 1, &FEBEX_EnergyLow);
    readNetInt("FEBEX_EnergyHigh", 0, 1, &FEBEX_EnergyHigh);
    readNetInt("FEBEX_BinSize", 0, 1, &FEBEX_BinSize);

    readNetString("FEBEX_DataBase", 0, 1, FEBEX_Data_Base);
    readNetString("FEBEX_DataInFile", 0, 1, FEBEX_DataIn_File);
    readNetString("FEBEX_DataOutFile", 0, 1, FEBEX_DataOut_File);
    
  printf ("febex_init_connection  %d %d %d %d %d \n", FEBEX_PexorEnable, FEBEX_Data_In_Enable, FEBEX_Data_Out_Enable, FEBEX_TestPatternMode, FEBEX_TokenMode);
  printf ("%d %d %d %d %d\n", FEBEX_MWDFormat, FEBEX_ReadOutDelay, FEBEX_EnergyLow, FEBEX_EnergyHigh, FEBEX_BinSize);
  printf ("FEBEX_DataBase %s:\n", FEBEX_Data_Base);
  printf ("FEBEX_DataIn_File %s:\n", FEBEX_DataIn_File);
  printf ("FEBEX_DataOutFile %s:\n", FEBEX_DataOut_File);

    
    if (FEBEX_PexorEnable == 1) {
  
      /* open pexor device */
      
       fd_pex = mbspex_open(PEXDEVNO);
       if (fd_pex < 0) {
            printf("ERROR>> open /dev/pexor%d \n", PEXDEVNO);
 //          return;
       }
    
       if(*dacq_Trace > 1) printf ("pexor device /dev/pexor%d opened  %d \n", PEXDEVNO, fd_pex);
    
    }
    
    if (FEBEX_Data_In_Enable == 1) {
      
      sprintf(fname,"%s/%s",FEBEX_Data_Base,FEBEX_DataIn_File);
      FEBEX_DataInHandle = fopen (fname,"r");
      if (FEBEX_DataInHandle < 0) {printf("ERROR>> open file %s (read) \n", fname);} else {printf ("In File = %s\n", fname);}
      
    }
    
    if (FEBEX_Data_Out_Enable == 1) {
      
//    see action in febex_go_connection
      
//      sprintf(fname,"%s/%s",FEBEX_Data_Base,FEBEX_DataOut_File);
//      FEBEX_DataOutHandle = fopen (fname,"w");
//      if (FEBEX_DataOutHandle < 0) {printf("ERROR>> open file %s (write) \n", fname);} else {printf ("Out File = %s\n", fname);}
      
    }
    
    readNetInt("FEBEX.Link0NumBoards", 0, 1 , &nslaves[0]); 
    printf ("%d, SFP id: %d  slaves %d\n",rc, 0, nslaves[0]);
    
    readNetInt("FEBEX.Link1NumBoards", 0, 1 , &nslaves[1]);
    printf ("%d, SFP id: %d  slaves %d\n",rc, 1, nslaves[1]);


    readNetInt("FEBEX.Link2NumBoards", 0, 1 , &nslaves[2]); 
    printf ("%d, SFP id: %d  slaves %d\n",rc, 2, nslaves[2]);

    readNetInt("FEBEX.Link3NumBoards", 0, 1 , &nslaves[3]); 
    printf ("%d, SFP id: %d  slaves %d\n",rc, 3, nslaves[3]);
    
    if (FEBEX_PexorEnable == 1) {

   for (sfp_channel=0; sfp_channel < nsfp; sfp_channel++)
   {
      if (nslaves[sfp_channel] != 0)
      {
	 sfp_phys_mem_base[sfp_channel] = (long) PEX_MEM_OFF + (long)(PEX_SFP_OFF * sfp_channel);
      
          if(*dacq_Trace > 1) printf ("SFP id: %d Pexor SFP physical memory base 0x%08x \n", sfp_channel, sfp_phys_mem_base[sfp_channel]);
      }
   }

    // map pipe memory
    //  these options for pipe memory
    len = (size_t) PCI_BAR0_SIZE;
    off = 0;
    off = (off_t) 0x40000000;
    int prot = PROT_WRITE | PROT_READ;
    int flags = MAP_SHARED | MAP_LOCKED;

    if(*dacq_Trace > 1) printf("map pipe at 0x%x, size: 0x%x, ...\n", off, len);
    if ((pa = (char *) mmap(NULL, len, prot, flags, fd_pex, off)) == MAP_FAILED) {
        if(*dacq_Trace > 1) printf("failed to mmap pipe at pex, return: 0x%x, %d \n", pa, pa);
        perror("mmap");
    } else {
        pe = pa + len - 1;
        if(*dacq_Trace > 1) printf("physical address:             0x%x \n", off);
        if(*dacq_Trace > 1) printf("size:                         0x%x \n", len);
        if(*dacq_Trace > 1) printf("first mapped virtual address: 0x%x \n", pa);
        if(*dacq_Trace > 1) printf("last  mapped virtual address: 0x%x \n", pe);
        pdat = (int *) pa;
    }
    
    l_status = mbspex_reset(fd_pex);
    if (l_status != 0) {
        if(*dacq_Trace > 1) printf("\n\nError %d in mbspex_reset . exiting...\n", l_status);
        febex_term_connection();
    }
    
    for (sfp_channel=0; sfp_channel< nsfp; sfp_channel++)
    {
      if (nslaves[sfp_channel] != 0)
      {
        l_status = mbspex_slave_init(fd_pex, sfp_channel, nslaves[sfp_channel]);
        if (l_status != 0) {
           if(*dacq_Trace > 1) printf("\n\nError %d in mbspex_slave_init for channel %x . exiting...\n", l_status, sfp_channel);
           febex_term_connection();
        }
        bufid[sfp_channel] = 0;
      }
    }
    
    }  /*  end of  if FEBEX_PexorEnable == 1   */
    
    FEBEX_FifoBufferPtr[module] = 0;    /*  reset */


}

void febex_readevent ()
{
  
    struct timespec timedelay;
    
    int l_status;
    int t_status;
    long l_memsize;
    long l_toksize;
    long token_status;

    long check_comm = 0, check_token = 0, check_sls = 0;
    int totalsize;
    int * pdata;
    
    int validdata = 0;
    
    int i, j;
    int module = 0;
    int rc;
    

/* do synchronous token request with direct dma*/
/*    returns output data in databuffer & databufferlen  */

	
	if (FEBEX_PexorEnable == 1) {
	  
      timedelay.tv_sec = 0;
      timedelay.tv_nsec = FEBEX_ReadOutDelay;
      nanosleep (&timedelay, NULL);    /*  5 millisec  */	  
    
           if(*dacq_Trace > 1) printf("**** Requesting token from  sfp 0x%x, bufid %d, tok_mode %d ...\n", sfp_channel, bufid[sfp_channel], tok_mode);

           if(*dacq_Trace > 1) printf("/dev/pexor %d; physical address: 0x%x \n", fd_pex, off);

           if(*dacq_Trace > 1) printf("first mapped virtual address: 0x%x \n", pa);
	   
	    t_status = mbspex_slave_rd(fd_pex, sfp_channel, module, REG_RST, (long *) &token_status);
            if(*dacq_Trace > 0) printf("\npre %x in mbspex_slave_rd for channel %d module %d token_status 0x%x\n",t_status, sfp_channel, module, token_status);
	    
	   dmasize = 0;
           l_status = mbspex_send_and_receive_tok(fd_pex, sfp_channel, tok_mode + bufid[sfp_channel], off, &dmasize, &check_comm, &check_token, &check_sls);
	        /* set target to begin of pipe memory */

           if (l_status != 0) {
              if(*dacq_Trace > 1) printf("\n\nError %x in mbspex_send_and_receive_tok for channel %d\n",l_status, sfp_channel);
           }
           
            if(*dacq_Trace > 0) printf("\nError %x in mbspex_send_and_receive_tok for channel=%d dmasize=0x%x check_token=0x%x\n",l_status, sfp_channel, dmasize, check_token); 
	    
	    t_status = mbspex_slave_rd(fd_pex, sfp_channel, module, REG_RST, (long *) &token_status);
            if(*dacq_Trace > 0) printf("\npost %x in mbspex_slave_rd for channel %d module %d token_status 0x%x\n",t_status, sfp_channel, module, token_status); 
	    
	    
	    if(*dacq_Trace > 0) printf("+++++> send and receive token\n");
            if(*dacq_Trace > 0) printf("sfp_channel = %d; l_status = %d\n",sfp_channel, l_status);
            if(*dacq_Trace > 0) printf("tok_mode = %d; bufid = %d; off = 0x%x; check_token = %d; slaves = %d\n", tok_mode, bufid[sfp_channel], off, check_token, check_sls);
	  
        }  /*  end of (if FEBEX_PexorEnable == 1)  */
	
       if (FEBEX_PexorEnable == 0 && FEBEX_Data_In_Enable == 1 && FEBEX_DataInHandle >= 0) {
	
//    read raw data previous written to disc file
	 
      timedelay.tv_sec = 0;
      timedelay.tv_nsec = FEBEX_ReadOutDelay * 100;
      nanosleep (&timedelay, NULL);    /* default -  1/50 sec  */
	 
	   validdata = 0;
	  
	  rc = fread(&totalsize, 1, sizeof(totalsize), FEBEX_DataInHandle);   /* length of data block  */
	  
	  if (rc == sizeof(totalsize)) {
	      rc = fread(databuffer, 1, totalsize, FEBEX_DataInHandle);  /*  read the data block  */
              if (rc == totalsize ) {
	         dmasize = totalsize;
	         databufferlen = totalsize/4;
		     
	         rc = fread(&eod, 1, sizeof(EOD), FEBEX_DataInHandle);	/*  the end marker  */
		 if (rc == sizeof(eod) && eod == EOD) {
		    validdata = 1;
	            if(*dacq_Trace > 1) printf("sfp_channel = %d; databufferlen = %d\n",sfp_channel,databufferlen);
		 }		 	
	      }
	  }
	  
	  if (validdata == 0) {
	     totalsize = 0;
	     databufferlen = 0;
	     
//	     if (rc == 0) {printf(">> end of file %s\n", fname);} 
	     if (rc != 0) {printf("ERROR>> %d reading file %s\n", rc, fname);}
	     
	     fclose(FEBEX_DataInHandle);
	  
             sprintf(fname,"%s/%s",FEBEX_Data_Base,FEBEX_DataIn_File);
             FEBEX_DataInHandle = fopen (fname,"r");
             if (FEBEX_DataInHandle < 0) {printf("ERROR>> open file %s (read) \n", fname);}
//             if (FEBEX_DataInHandle >= 0} {printf ("In File = %s\n", fname);}
	  }
          
       }   /*  end of (FEBEX_PexorEnable == 0)  */
       
       if (FEBEX_PexorEnable == 1) { 
       
       bufid[sfp_channel] = (bufid[sfp_channel] + 1) & 1;     /* update buffer toggle flag */

       if(*dacq_Trace > 1) printf("dma size on return:  0x%x = %d bytes \n", dmasize, dmasize);
       
       if (dmasize > DMASIZE) {
	   printf("ERROR>> returned data buffer size invalid %d > %d\n", dmasize, DMASIZE);
	   totalsize = 0; 
       } else {
	   totalsize = dmasize;
       }
      
       if (*dacq_Trace > 1) printf("Using total data size = %d bytes\n", totalsize);

       for (j=0;j<(totalsize+3)/4;j++) {  /* copy data buffer */
	     databuffer[j] = pdat[j];
       }
       databufferlen = (totalsize+3)/4;   /*   ints   */
       
      }   /*  end of  if FEBEX_PexorEnable == 1  */

      if (FEBEX_Data_Out_Enable == 1 && FEBEX_DataOutHandle >= 0 && totalsize > 0) {

      rc = fwrite (&totalsize, 1, sizeof (totalsize), FEBEX_DataOutHandle);
      if (rc < 0) {
	    printf("ERROR>> write file %s \n", fname);
	    *dacq_State = STOPPED;
      }
      
      rc = fwrite (databuffer, 1, totalsize, FEBEX_DataOutHandle);
      if (rc < 0) {
	    printf("ERROR>> write file %s \n", fname);
	    *dacq_State = STOPPED;
      }
      
      rc = fwrite (&EOD, 1, sizeof (EOD), FEBEX_DataOutHandle);
      if (rc < 0) {
	    printf("ERROR>> write file %s \n", fname);
	    *dacq_State = STOPPED;
      }
      
    }
    
    
        febexfifobufferlength = FEBEXFIFOBUFFERLENGTH;
        if ((totalsize+3)/4 > febexfifobufferlength) {febexfifobufferlength = FEBEXFIFOBUFFERLENGTH;} else {febexfifobufferlength = (totalsize+3)/4;}
       
        FEBEX_FifoBuffer[module][0] = 0xa5a5a5a5;
	
        FEBEX_FifoBuffer[module][1] = totalsize+4;
	
	FEBEX_FifoBufferPtr[module] = 2;
       
        for (i=0; i<febexfifobufferlength; i++) {
 	   FEBEX_FifoBuffer[module][FEBEX_FifoBufferPtr[module]] = databuffer[i];
	   FEBEX_FifoBufferPtr[module]++;
	}
        FEBEX_FifoBuffer[module][FEBEX_FifoBufferPtr[module]] = 0x5a5a5a5a;
	
	   if (databuffer[1] > 8) {
//             if(*dacq_Trace > 0) printf("data = 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x \n", databuffer[0], databuffer[1], databuffer[2], databuffer[3],  databuffer[4], databuffer[5], databuffer[6], databuffer[7], databuffer[8], databuffer[9]);
           } else {
	       if (databuffer[1] == 0) { 
	          if (*dacq_Trace > 0) printf("no data!!\n");
	       } else {
                  if(*dacq_Trace > 0) printf("data?? = 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x 0x%08x\n", databuffer[0], databuffer[1], databuffer[2], databuffer[3],  databuffer[4], databuffer[5]);
	       }
	   }
}

void  febex_readdata() {   /*  called from FEBEX_interrupt in interrupt.c  */

    int i;
    int module = 0;
    
    struct timespec timedelay;
   
    evntbuffer = (unsigned int *) EventBufferAddr;   /* eventual output */
    evntbufferlen = 0;

    sigtrapset ("Error reading FEBEX modules");
    
    if (((*dacq_State) & 255) == GOING) {
      
    for (sfp_channel=0; sfp_channel < nsfp; sfp_channel++)
    {
      if (nslaves[sfp_channel] != 0)
      {
        
         if (FEBEX_Do_DataRead == 1) {febex_readevent();}
         
//         if (bufid[sfp_channel] == 0) {
	 
	 if (FEBEX_Do_DataFormat == 1)  {
	   
	    febex_formatevent();
	 
//    event now in event data buffer
//    We handle here rather than FEBEX_interrupt to avoid "events" exceeding 64K bytes
	 
	    sigtrapset ("Error during histogramming");
	    FEBEXhistogram_data();  /* possibly histogram the input data */
	 
	    sigtrapset ("error in xfer procedures");
	    xfer_event(0);         /* possibly send for storage */
	 }
	 
//	 }
	
     }      // if (nslaves[sfp_channel] != 0)
      
    }        // for (sfp_channel=0; sfp_channel < nsfp; sfp_channel++)
	
    } else {   /* STOPPED  */
    
      timedelay.tv_sec = 0;
      timedelay.tv_nsec = 10000000;
      nanosleep (&timedelay, NULL);    /*  10 millisec  */
      
    }
    
if(*dacq_Trace > 1)printf ("10 End of FEBEX_readdata %d\n", EventBufferLen);
    
    sigtrapset ("Error detected during data acquisition");

    return;

}




