/*

    code insert to handle VME interrupt setup and VME DMA transfer setup
    via the vmeint device driver
    
*/

#include <sys/mman.h>


void VME_int_setup ()
{

#ifndef LINUX
   vmeintcmd_t command;
#else
   int vector, level, flags;
   static struct sigevent event;
   vme_master_handle_t window_handle = NULL;
#endif

   int i;

/*    establish virtual addresses for the 3 VME address spaces   */

#ifdef LINUX
	if (vme_init(&VME_bus_handle) != 0) {
              perror ("Error in vme_init:");
              sprintf (message, "Unable to initialise VME");
              reportandquit();
	}
#endif

   readNetInt("VMEA32Size", 0, 1, &VMEA32Size);
   readNetInt("VMEA32Base", 0, 1, &VMEA32VMEBase);
   if (VMEA32Size >= 0 && VMEA32VMEBase >= 0) {
       if (VMEA32Size == 0) VMEA32Size = 0x5800000;
#ifndef LINUX
       VMEA32Base = (int)vmeMapA32(VMEA32VMEBase,VMEA32Size);
#else

	if (vme_master_window_create(VME_bus_handle, &window_handle,
                   VMEA32VMEBase,  VME_A32UD, VMEA32Size, VME_CTL_MAX_DW_32, NULL) != 0) {
              perror ("Error in vme_master_window_create:");
              sprintf (message, "Unable to create A32 VME window");
              reportandquit();
	}

	VMEA32Base = (int)vme_master_window_map(VME_bus_handle, window_handle, 0);
	if (VMEA32Base == (int)NULL) {
              perror ("Error in vme_master_window_map:");
              sprintf (message, "Unable to map VME A32 window");
              reportandquit();
	}
#endif
       printf("Mapped VME A32 address 0x%x to virtual address 0x%x for 0x%x bytes\n", VMEA32VMEBase, VMEA32Base, VMEA32Size);
   }

   readNetInt("VMEA24Size", 0, 1, &VMEA24Size);
   readNetInt("VMEA24Base", 0, 1, &VMEA24VMEBase);
   if (VMEA24Size >= 0 && VMEA24VMEBase >= 0) {
       if (VMEA24Size == 0) VMEA24Size = 0x1000000;
#ifndef LINUX
       VMEA24Base = (int)vmeMapA24(VMEA24VMEBase,VMEA24Size);
#else

	if (vme_master_window_create(VME_bus_handle, &window_handle,
                   VMEA24VMEBase,  VME_A24UD, VMEA24Size, VME_CTL_MAX_DW_32, NULL) != 0) {
              perror ("Error in vme_master_window_create:");
              sprintf (message, "Unable to create VME A24 window");
              reportandquit();
	}

	VMEA24Base = (int)vme_master_window_map(VME_bus_handle, window_handle, 0);
	if (VMEA24Base == (int)NULL) {
              perror ("Error in vme_master_window_map:");
              sprintf (message, "Unable to map VME A24 window");
              reportandquit();
	}

#endif
       printf("Mapped VME A24 address 0x%x to virtual address 0x%x for 0x%x bytes\n", VMEA24VMEBase, VMEA24Base, VMEA24Size);
   }

   readNetInt("VMEA16Size", 0, 1, &VMEA16Size);
   readNetInt("VMEA16Base", 0, 1, &VMEA16VMEBase);
   if (VMEA16Size >= 0 && VMEA16VMEBase >= 0) {
       if (VMEA16Size == 0) VMEA16Size = 0x10000;
#ifndef LINUX
       VMEA16Base = (int)vmeMapA16(VMEA16VMEBase,VMEA16Size);
#else

	if (vme_master_window_create(VME_bus_handle, &window_handle,
                   VMEA16VMEBase,  VME_A16U, VMEA16Size, VME_CTL_MAX_DW_16, NULL) != 0) {
              perror ("Error in vme_master_window_create:");
              sprintf (message, "Unable to create VME A16 window");
              reportandquit();
	}

	VMEA16Base = (int)vme_master_window_map(VME_bus_handle, window_handle, 0);
	if (VMEA16Base == (int)NULL) {
              perror ("Error in vme_master_window_map:");
              sprintf (message, "Unable to map VME A16 window");
              reportandquit();
	}

#endif
       printf("Mapped VME A16 address 0x%x to virtual address 0x%x for 0x%x bytes\n", VMEA16VMEBase, VMEA16Base, VMEA16Size);
   }

/*    setup the VME interrupt handler  */

#ifdef LYNXOS
     taskDelay(60);     /*  1 sec  */
#else
     usleep (1000000);
#endif

#ifndef LINUX           /*  for Linux the VME device is openned in the VME address mapping above  */

   vfd = open (VMEINT_DEVICE, O_RDWR);
   if (vfd == -1) {
      sprintf (message, "Unable to open vme inferface device");
      reportandquit();
   }
#endif

   if (DACQ_INT_VECTORS == 0) {return;}    /*  no VME interrupts required */

#ifndef LINUX

   printf("Installing VME interrupt handlers\n");

/*    inform the handler how to tell us about interrupts  (our PID and the signal to use)   */

   command.readit = 0;
   command.value = getpid();
   if (ioctl(vfd,VME_INT_PID, &command) == -1) {
      sprintf (message, "Unable to send ioctl to vme interrupt device for PID");
      close (vfd);
      reportandquit();
   }

   command.readit = 0;
   command.value = SIGINT;
   if (ioctl(vfd,VME_INT_SIG, &command) == -1) {
      sprintf (message, "Unable to send ioctl to vme interrupt device for SIG");
      close (vfd);
      reportandquit();
   }

/*    and setup handlers for the n interrupt vectors possible from the hardware module  */
/*    for the D2VB/F2VB module we use 8 interrupt vectors   */

  for(i=0; i<DACQ_INT_VECTORS; i++)
  {
       command.readit=0;
       command.value=DACQ_INT_BASE + i;
       if (ioctl(vfd,VME_INT_VECTOR, &command) == -1) {
          sprintf (message, "Unable to send ioctl to vme interrupt device for vector %d", command.value);
          close (vfd);
          reportandquit();
       }
   }

#else

      /* Set up sigevent structure  - we already have a signal handler for SIGINT  */

      event.sigev_signo = SIGINT;
      event.sigev_notify = SIGEV_SIGNAL;
      event.sigev_value.sival_int = 0;

/*    and setup handlers for the n interrupt vectors possible from the hardware module  */
/*    for the D2VB/F2VB module we use 8 interrupt vectors   */

      level = DACQ_INT_LEVEL;    /*     we use interrupt level 4 normally */
      flags = VME_INTERRUPT_SIGEVENT;

      for(i=0; i<DACQ_INT_VECTORS; i++)
      {
          vector = DACQ_INT_BASE + i;
          printf ("Installing interrupt handler, level = %d, vector = 0x%x, flags = 0x%x\n", level, vector, flags);

          if (vme_interrupt_attach (VME_bus_handle, &VME_interrupt_handle,  level, vector, flags, (void *) &event) < 0) {
              perror ("Error in vme_interrupt_attach:");
              sprintf (message, "Unable to install VME interrupt handler");
              reportandquit();
          }
      }

#endif

      printf("Done\n");

}

void VME_dma_setup()
{
   int shm;
   char * BufPtr;

   if (DACQ_BufferSize == 0) {return;}    /*  no DMA buffer required */

   readNetInt("DMAInhibit", 0, 1, &DMAInhibit);

/*
      Note - if DMAInhibit is true we must still obtain the buffer because it is used in VME_xfer
      Also the address and data options are required for the program loop
*/

    if (DMAInhibit != 0) {printf("DMA is inhibited - data readout will be via program loop\n");}
 
   printf("Setting up DMA access to Hardware Buffer RAM\n");

/*    now setup DMA for reading the hardware memory  */

#if defined MVME5500 
{                                          /*    This is for the MVME5500 with LynxOS   */
    vmeintmem_t    vme_mem = {0, 0};
    vme_mem.data = DACQ_BufferSize;
    if (ioctl(vfd, VME_INT_ALLOC_CMEM, &vme_mem) == -1) {
                sprintf(message, "Unable to allocate buffer memory. (Errno %d - %s)", errno, strerror(errno));
                reportandquit();
    }

    shm = open("/dev/mem", O_RDWR);
    if ( shm == -1 ) {
        sprintf(message, "Unable to open memory device (/dev/mem). (Errno %d - %s)", errno, strerror(errno));
                reportandquit();
    }
    BufPtr = mmap(NULL, DACQ_BufferSize, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_CONTIG, shm, vme_mem.addr);
    if ( BufPtr == MAP_FAILED ) {
                close (shm);
                sprintf(message, "Unable to mmap DMA buffer. (Errno %d - %s)", errno, strerror(errno));
                reportandquit();
    }
    if ( mlock(BufPtr, DACQ_BufferSize) ) {
        sprintf(message, "Merge: Unable to mlock allocated buffer. (Errno %d - %s)", errno, strerror(errno));
        (void) munmap(BufPtr, DACQ_BufferSize);
                reportandquit();
    }
    close (shm);

#ifdef USEINT
    DACQ_Kernel_InputBuffer = (unsigned int *)vme_mem.data;
#endif

#ifdef USESHORT
    DACQ_Kernel_InputBuffer = (unsigned short *)vme_mem.data;
#endif

}

#else

#ifndef LINUX

/* Use shm_open()/mmap() to allocate some contiguous physical memory */

    shm = shm_open(BUF_NAME, O_RDWR | O_CREAT, 0666);
    if ( shm == -1 ) {
        sprintf(message, "Unable to shm_open DMA buffer. (Errno %d - %s)", errno, strerror(errno));
        reportandquit();
    } else {
        if ( ftruncate(shm, DACQ_BufferSize) ) {
            close (vfd);
            close (shm);
            sprintf(message, "Unable to ftruncate DMA buffer. (Errno %d - %s)", errno, strerror(errno));
            reportandquit();
        } else {
            BufPtr = mmap(NULL, DACQ_BufferSize, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_CONTIG, shm, 0);
            if ( BufPtr == MAP_FAILED ) {
                (void) shm_unlink(BUF_NAME);
                close (shm);
                sprintf(message, "Unable to mmap DMA buffer. (Errno %d - %s)", errno, strerror(errno));
                reportandquit();
            }
        }
    }
    close (shm);
#else
                                                       /*  Linux for MVME5500  */

    if (vme_dma_buffer_create(VME_bus_handle, &VME_dma_handle, DACQ_BufferSize, 0, NULL) != 0) {
              perror ("Error in vme_dma_buffer_create:");
              sprintf (message, "Unable to create DMA buffer");
              reportandquit();
    }

    BufPtr = (char *)vme_dma_buffer_map(VME_bus_handle, VME_dma_handle, 0); 
    if (BufPtr == NULL) {
        perror ("Error in vme_dma_buffer_map:");
        sprintf(message, "Unable to map DMA buffer");
        reportandquit();
    }

#endif

#ifdef USEINT
    DACQ_Kernel_InputBuffer = (unsigned int *)BufPtr;
#endif

#ifdef USESHORT
    DACQ_Kernel_InputBuffer = (unsigned short *)BufPtr;
#endif

#endif

#ifdef USEINT
    DACQ_InputBuffer = (unsigned int *)BufPtr;
#endif

#ifdef USESHORT
    DACQ_InputBuffer = (unsigned short *)BufPtr;
#endif

    PCIAddressOffset= 0;

    printf ("DMA buffer of size %d allocated at 0x%08x\n", DACQ_BufferSize, DACQ_InputBuffer);

/*    Zero the buffer */

    printf ("Zeroing DMA buffer\n");
    memset(DACQ_InputBuffer, 0, DACQ_BufferSize);
    printf("Done\n");
    
/*    setup DMA mode   */

     readNetInt("DMAFlags_vdw", 0, 1, &ScratchNetInt);
     switch (ScratchNetInt) {
         case  0:         /*  undefined - use builtin default */
             DMA_flags.vdw = D_VDW_32;
             printf("VMEbus maximum datawidth (vdw) = %d bits\n", VDW);
             break;
         case  16:
             DMA_flags.vdw = D_VDW_16;
             printf("VMEbus maximum datawidth (vdw) = %d bits\n", ScratchNetInt);
             break;
         case  32:
             DMA_flags.vdw = D_VDW_32;
             printf("VMEbus maximum datawidth (vdw) = %d bits\n", ScratchNetInt);
             break;
         case  64:
             DMA_flags.vdw = D_VDW_64;
             printf("VMEbus maximum datawidth (vdw) = %d bits\n", ScratchNetInt);
             break; 
         default:
             printf("Unknown VDW option (%d) - using program default = %d bits\n", ScratchNetInt, VDW);
             break;
     }
     
     readNetInt("DMAFlags_vas", 0, 1, &ScratchNetInt);
     switch (ScratchNetInt) {
         case  0:         /*  undefined - use builtin default */
             DMA_flags.vas = D_VAS_32;
             printf("VMEbus address space (vas) = %d bits\n", VAS);
             break;
         case  24:
             DMA_flags.vas = D_VAS_24;
             printf("VMEbus address space (vas) = %d bits\n", ScratchNetInt);
             break;
         case  32:
             DMA_flags.vas = D_VAS_32;
             printf("VMEbus address space (vas) = %d bits\n", ScratchNetInt);
             break;
         default:
             printf("Unknown VAS option (%d) - using program default = %d bits\n", ScratchNetInt, VAS);
             break;
     }
     
     readNetInt("DMAFlags_vct", 0, 1, &ScratchNetInt);
     switch (ScratchNetInt) {
         case  0:         /*  undefined - use builtin default */
             DMA_flags.vct = VCT_SC;
             printf("VMEbus cycle type (vct) = %s\n", VCT);
             break;
         case  1:
             DMA_flags.vct = VCT_SC;
             printf("VMEbus cycle type (vct) = SC\n");
             break;
         case  2:
             DMA_flags.vct = VCT_SCBT;
             printf("VMEbus cycle type (vct) = BLT\n");
             break;
         default:
             printf("Unknown VCT option (%d) - using program default = %s\n", ScratchNetInt, VCT);
             break;
     }
                      
}

void VME_xfer(void *addr, unsigned int len)
{
/*    move data from hardware buffer to  processor RAM buffer   */

#ifdef LINUX
    unsigned long vme_addr;
    unsigned long nelem;
    vme_addr_mod_t  addr_mod;
    vme_dwidth_t dwidth;
    unsigned long flags;
    int status;
#endif

    volatile short * from16;
    short * to16;
    volatile long * from32;
    long * to32;
    int length;

    VME_dma_errcode = 0;

    if (DMAInhibit != 0) {

      if (DMA_flags.vdw == D_VDW_16) {
          if (DMA_flags.vas == D_VAS_24) {
              from16 = (volatile short *) ((long)addr - VMEA24VMEBase + VMEA24Base);
          } else {
              from16 = (volatile short *) ((long)addr - VMEA32VMEBase + VMEA32Base);
          }
          to16 = (short *) DACQ_InputBuffer;
          length = len >> 1;
          while (length-- > 0) {*to16++ = *from16++;}
      } else {
          if (DMA_flags.vas == D_VAS_24) {
              from32 = (volatile long *) ((long)addr - VMEA24VMEBase + VMEA24Base);
          } else {
              from32 = (volatile long *) ((long)addr - VMEA32VMEBase + VMEA32Base);
          }
          to32 = (long *) DACQ_InputBuffer;
          length = len >> 2;
          while (length-- > 0) {*to32++ = *from32++;}
      }

          return;
    }

      if (DMA_flags.vdw == D_VDW_16) {
          to16 = (short *) DACQ_InputBuffer;
          length = DACQ_BufferSize >> 1;
//          while (length--) {*to16++ = 0;}        /*    zero input buffer - useful for debugging  */
      } else {
          to32 = (long *) DACQ_InputBuffer;      //    use the whole buffer - not just in use bit
          length = DACQ_BufferSize >> 2;
//          while (length--) {*to32++ = 0;}        /*    zero input buffer - useful for debugging  */
      }

#ifndef LINUX

/* Other members of struct are initialised with constants in DACQ_VME.h */
    DMA_flags.pci_addr = (int)DACQ_Kernel_InputBuffer + PCIAddressOffset;
    DMA_flags.vme_addr = (long) addr; /* ++++ really should be long* */
    DMA_flags.byte_cnt = ((len + 7)/8)*8;   /* round to 8 bytes for Tundra Universe */

    if (ioctl(vfd,VME_INT_DMA_PGM_VIRT, &DMA_flags) == -1) {
        sprintf(message, "ioctl DMA_PGM failed. (Errno %d - %s)", errno, strerror(errno));
        reportandquit();
    }

    if(*dacq_Trace == 1)
          printf("Setting DMA GO for %d bytes from 0x%08x (VME) to 0x%08x (RAM)\n",
             DMA_flags.byte_cnt, DMA_flags.vme_addr, DMA_flags.pci_addr);

/*    set DMA going   */
    if (ioctl(vfd,VME_INT_DMA_GO, NULL) == -1) {
        sprintf(message, "ioctl DMA_GO failed. (Errno %d - %s)", errno, strerror(errno));
        reportandquit();
    }

/*    wait for completion  and invalidate cache for this region */
    DMA_flags.pci_addr = (int)DACQ_Kernel_InputBuffer + PCIAddressOffset;
    DMA_flags.byte_cnt = len;
    if (ioctl(vfd,VME_INT_DMA_WAIT, &DMA_flags) == -1) {
        sprintf(message, "ioctl DMA_WAIT failed. (Errno %d - %s)", errno, strerror(errno));
        reportandquit();
    }

#else

     vme_addr = (unsigned long)addr;

     switch (DMA_flags.vdw) {
         case  D_VDW_16:         /*  D16 */
             dwidth = 2;
             nelem = (len+1)/2;
             nelem = ((nelem + 3)/4)*4;    /* round to 8 bytes for Tundra Universe */
             flags = VME_DMA_DW_16;
             break;
         case  D_VDW_32:         /*  D32 */
             dwidth = 4;
             nelem = (len+3)/4;
             nelem = ((nelem + 1)/2)*2;    /* round to 8 bytes for Tundra Universe */
             flags = VME_DMA_DW_32;
             break;
         case  D_VDW_64:         /*  D64  */
             dwidth = 8;
             nelem = (len+7)/8;
             flags = VME_DMA_DW_64;
             break;
     }

     switch (DMA_flags.vas) {
         case  D_VAS_16:         /*  A16 */
             addr_mod = VME_A16U;    /* 0x29 */
             break;
         case  D_VAS_24:         /*  A24 */
             addr_mod = VME_A24UB;        /* 0x3B */
             break;
         case  D_VAS_32:         /*  A32  */
             switch (DMA_flags.vdw) {
                case  D_VDW_64:         /*  D64  */
                   addr_mod = VME_A32UMB;      /*  0x08  */
                   break;
                default:
                   addr_mod = VME_A32UB;      /*  0x0B  */
                   break;
             }
             break;
     }

    if(*dacq_Trace == 1) {
         printf("DMA request: bus_handle 0x%x dma_handle 0x%x flags 0x%x\n",VME_bus_handle, VME_dma_handle, flags);
         printf("DMA request: 0x%x 0x%x 0x%x 0x%x\n",vme_addr, addr_mod, nelem, dwidth);
    }

    if (vme_dma_read(VME_bus_handle, VME_dma_handle, PCIAddressOffset, vme_addr, addr_mod, nelem * dwidth, flags) != 0)
    {
/*
        perror("Error in vme_dma_read:");
        sprintf(message, "DMA failed");
        report();
        sprintf(message,"DMA request: bus_handle 0x%x dma_handle 0x%x flags 0x%x",VME_bus_handle, VME_dma_handle, flags);
        report();
        sprintf(message,"DMA request: 0x%x 0x%x 0x%x 0x%x",vme_addr, addr_mod, nelem, dwidth);
        report();
*/
        VME_dma_errcode = -1;
    } 

#endif

    if(*dacq_Trace == 1) printf("DMA completed\n");

}
