


/*
    This code insert is the Hit/Rate histogram thread
*/


#define MAXCHANNELS 0x10000


#ifndef WIN32
#include <pthread.h>
#endif

#ifdef LYNXOS
#include <time.h>
#include <resource.h>
#else
#ifdef WIN32
#else
#include <unistd.h>
#endif
#endif



/* --------------------------------------------------------------------- */

#ifndef WIN32
void *DACQ_HitHist(void * arg)
#else
DWORD WINAPI DACQ_HitHist(LPVOID arg)
#endif
{
   int i;
   int *delay;
   unsigned int *Hitmemaddr, *Ratememaddr;
   unsigned int Hitmemlen, Ratememlen;
   unsigned int memlen;
   int *dacq_State;
   unsigned int Rate[MAXCHANNELS+1];
   unsigned int Count[MAXCHANNELS+1];
   unsigned int Hit[MAXCHANNELS+1];

#ifdef LYNXOS
   int id, prty1, prty2;
#endif

#ifdef WIN32
    char msgBuf[BUF_SIZE];
    DWORD dwChars;
#endif

#ifndef WIN32
     printf("Hit/Rate creation thread starting\n");
#else
    sprintf(msgBuf,"Hit/Rate creation thread starting\n");
    WriteConsole(hStdout, msgBuf, strlen(msgBuf), &dwChars, NULL);
#endif

    signal_block();       //    mask off signals handled by the main thread

//   adjust priority - run just above main data acquisition program

#ifdef LYNXOS
          id = BUILDPID(getpid(), getstid());
          prty1 = getpriority(PRIO_PROCESS, id);
          (void) setpriority(PRIO_PROCESS, id, prty1 + 2);
          prty2 = getpriority(PRIO_PROCESS, id);
          printf("Hit/Rate thread priority changed from %d to %d\n", prty1, prty2);
#else
#ifdef WIN32
    if (!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST))  ErrorHandler(TEXT("SetThreadPriority"));
#else
     nice(-2);
#endif
#endif

   dacq_State = (int *)getNetIntMem("DACQ_State");   /* data acquisition state */
   delay = (int *)getNetIntMem("DACQ_StatsTime");
   
//    initialise the Hit and Rate histograms used for statistics rates

   Hitmemaddr=Ratememaddr=NULL;
   Hitmemlen=Ratememlen=0;
   
   for(i=0;i<=MAXCHANNELS;i++) {Count[i]=Rate[i]=0;}

  for (;;)
  {

#ifdef LYNXOS
     taskDelay((*delay)*60);   /*  delay units are seconds */
#else
#ifdef WIN32
     Sleep((*delay)*1000);
#else
     usleep ((*delay)*1000000);
#endif
#endif

     if (*dacq_State != GOING) {

         if (Hitmemaddr != NULL)
              (void) sasReleaseSpecMem("Hit", (unsigned char *)Hitmemaddr);
         Hitmemaddr=NULL;
         Hitmemlen=0;

         if (Ratememaddr != NULL)
             (void) sasReleaseSpecMem("HitRate", (unsigned char *)Ratememaddr);
         Ratememaddr=NULL;
         Ratememlen=0;

         continue;
     }

     if (Hitmemaddr == NULL) {
         Hitmemaddr = (unsigned int *) sasGetSpecMem("Hit");
         if (Hitmemaddr != NULL)
              Hitmemlen  = (unsigned int) (sasGetSpecLen("Hit") >> 2);
     }
     
     if (Ratememaddr == NULL) {
         Ratememaddr = (unsigned int *) sasGetSpecMem("HitRate");
         if (Ratememaddr != NULL)
             Ratememlen  = (unsigned int) (sasGetSpecLen("HitRate") >> 2);
     }
    
     if (Hitmemlen != Ratememlen) {continue;}

     memlen = Hitmemlen;
     
//    update data from Hit histogram

     if (Hitmemaddr != NULL) {
         for(i=0;i<(int)memlen;i++) {Hit[i] = Hitmemaddr[i];}  // current counters
     }

//    update rates
     
     for(i=0;i<(int)memlen;i++)  {Rate[i]=(Hit[i]-Count[i])/(*delay); Count[i]=Hit[i];}
     
     if (Ratememaddr != NULL) {
         for(i=0;i<(int)memlen;i++) {Ratememaddr[i] = Rate[i];}
     }
  }
}


/* --------------------------------------------------------------------- */

void DACQ_InitHitHist()
{

#ifndef WIN32
   pthread_t ThreadID;
   pthread_attr_t ThreadAttr;
   int err;

/*    create a thread for the Hit/Rate creation  */

       (void) pthread_attr_init(&ThreadAttr);
       (void) pthread_attr_setdetachstate(&ThreadAttr,PTHREAD_CREATE_DETACHED);

       err = pthread_create(&ThreadID, &ThreadAttr, DACQ_HitHist, NULL);
       if (err != 0) {
           perror("Error in pthread_create:");
           printf("Failed to start Hit/Rate thread: This must be fixed\n");
           exit(16);
       }

     printf("Hit/Rate creation thread created\n");

#else

    char msgBuf[BUF_SIZE];
    DWORD dwChars;

         hThreadArray[HITRATE] = CreateThread(
              NULL,                         // default security attributes
              0,                            // use default stack size
              DACQ_HitHist,                   // thread function name
              NULL,                         // argument to thread function
              0,                            // use default creation flags
              &dwThreadIdArray[HITRATE]);                        // returns the thread identifier

         if (hThreadArray[HITRATE] == NULL) {
             ErrorHandler(TEXT("CreateThread for HitHist"));
             ExitProcess(1);
         }

      sprintf(msgBuf, "Hit/Rate creation thread created (tid=%d)\n", dwThreadIdArray[HITRATE]);
    WriteConsole(hStdout, msgBuf, strlen(msgBuf), &dwChars, NULL);

         ThreadInUse[HITRATE] = 1;
         NumThreads++;

#endif


}

/* --------------------------------------------------------------------- */

