


/*

    code insert which provides the  task signal handling 
       This excludes support for Windows which is in the files WIN32signals
    
*/



//    the (trival) POSIX signal handlers

#if defined SAGE_SIGNALS
void sigalrm_received (int signo)
{
    sigalrm_occurred++;
    signal_occurred++;
}

void sigio_received (int signo)
{
    sigio_occurred++;
    signal_occurred++;
}
#endif
void sigint_received (int signo)
{
    sigint_occurred++;
    signal_occurred++;
}
void sigusr1_received (int signo)
{
    sigusr1_occurred++;
    signal_occurred++;
}
void sigusr2_received (int signo)
{
    sigusr2_occurred++;
    signal_occurred++;
}


#if defined SAGE_SIGNALS
static struct itimerspec starti = {{0,0}, {0,0}};
static struct itimerspec stopi  = {{0,0}, {0,0}};
#endif

void cond_wait_for_signal ()
{
   sigset_t newsignalSet;
   sigset_t oldsignalSet;

   if (signal_occurred == 0) 
   {
       sigemptyset(&newsignalSet);
       sigaddset(&newsignalSet, SIGINT);
#if defined SAGE_SIGNALS
       sigaddset(&newsignalSet, SIGIO);
       sigaddset(&newsignalSet, SIGALRM);
#endif
       sigaddset(&newsignalSet, SIGUSR1);
       sigaddset(&newsignalSet, SIGUSR2);
       sigprocmask(SIG_BLOCK, &newsignalSet, &oldsignalSet);

#ifdef SAGE_SIGNALS
       if (timer_settime(timerid, 0, &starti, NULL))  {
          perror("Unable to arm timeout timer - This must be fixed - exiting.");
          exit(1);
       }
#endif
   }
    
   if (signal_occurred == 0)
   {
       sigdelset(&oldsignalSet, SIGINT);
#if defined SAGE_SIGNALS
       sigdelset(&oldsignalSet, SIGIO);
       sigdelset(&oldsignalSet, SIGALRM);
#endif
       sigdelset(&oldsignalSet, SIGUSR1);
       sigdelset(&oldsignalSet, SIGUSR2);
       (void)sigsuspend(&oldsignalSet);
   }

   signal_occurred--;

#if defined SAGE_SIGNALS
   if (sigalrm_occurred > 0)
   {
           sigalrm_occurred--;
           Signal =SIGALRM;
           return;
   }
   if (sigio_occurred > 0)
   {
           sigio_occurred--;
           Signal =SIGIO;
           return;
   }
#endif
   if (sigusr1_occurred > 0)
   {
           sigusr1_occurred--;
           Signal =SIGUSR1;
           return;
   }
   if (sigusr2_occurred > 0)
   {
           sigusr2_occurred--;
           Signal =SIGUSR2;
           return;
   }
   if (sigint_occurred > 0)
   {
           sigint_occurred--;
           Signal =SIGINT;
           return;
   }
   Signal =  0;    /* unexpected event  */

#if defined SAGE_SIGNALS
   if (timer_settime(timerid, 0, &stopi, NULL))  {
      perror("Unable to disarm timeout timer - This must be fixed - exiting.");
      exit(1);
   }
#endif

}

void signal_block ()
{
   sigset_t signalSet;

//    mask off signals handled by the main thread

   sigemptyset(&signalSet);
   sigaddset(&signalSet, SIGINT);
#if defined SAGE_SIGNALS
   sigaddset(&signalSet, SIGIO);
   sigaddset(&signalSet, SIGALRM);
#endif
   sigaddset(&signalSet, SIGUSR1);
   sigaddset(&signalSet, SIGUSR2);
   sigprocmask(SIG_BLOCK, &signalSet, NULL);
}

/* --------------------------------------------------------------------- */

void signal_setup ()
{
   int i;
   struct sigaction sa;

   printf ("Installing signal handlers\n");

/*     write our PID into the netint for use by the Sig interface   */

/* Also done elsewhere for Collate to give control of multiple program instances */
    assignNetInt("PID_EXEC", 0, 1, getpid());
//    install required signal handlers

//    By default trap everything  (mostly)

   for (i=SIGHUP; i<SIGRTMIN; i++) {
      if (i==SIGTERM) continue;
      if (i==SIGWINCH) continue;

      sa.sa_handler = sigxxx_received;
      sigemptyset(&sa.sa_mask);
      sa.sa_flags=0;
      (void) sigaction(i, &sa, NULL);
   }

   signal_occurred=0;
   sigusr1_occurred=0;
   sa.sa_handler = sigusr1_received;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=0;

       if (sigaction(SIGUSR1, &sa, NULL))  {
          sprintf(message, "Unable to setup signal handler for SIGUSR1 - This must be fixed - exiting.");
          reportandquit();
       }

   sigusr2_occurred=0;
   sa.sa_handler = sigusr2_received;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=0;

       if (sigaction(SIGUSR2, &sa, NULL))  {
          sprintf(message, "Unable to setup signal handler for SIGUSR2 - This must be fixed - exiting.");
          reportandquit();
       }

   sigint_occurred=0;
   sa.sa_handler = sigint_received;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=0;

       if (sigaction(SIGINT, &sa, NULL))  {
          sprintf(message, "Unable to setup signal handler for SIGINT - This must be fixed - exiting.");
          reportandquit();
       }

#if defined SAGE_SIGNALS
   sigio_occurred=0;
   sa.sa_handler = sigio_received;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=0;

       if (sigaction(SIGIO, &sa, NULL))  {
          sprintf(message, "Unable to setup signal handler for SIGIO - This must be fixed - exiting.");
          reportandquit();
       }

   sigalrm_occurred=0;
   sa.sa_handler = sigalrm_received;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=0;

       if (sigaction(SIGALRM, &sa, NULL))  {
          sprintf(message, "Unable to setup signal handler for SIGALRM - This must be fixed - exiting.");
          reportandquit();
       }

   starti.it_value.tv_nsec = 1000000;
   if (timer_create(CLOCK_REALTIME, NULL, &timerid))  {
      perror("Unable to create timer - This must be fixed - exiting.");
      exit(1);
   }
#endif

/*    and finally a handler for SIGBUS errors to trap VME bus timeouts   */

   sa.sa_handler = sigbus_received;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=0;

       if (sigaction(SIGBUS, &sa, NULL))  {
          sprintf(message, "Unable to setup signal handler for SIGBUS - This must be fixed - exiting.");
          reportandquit();
       }

/*    plus a handler for SIGSEGV errors    */

   sa.sa_handler = sigsegv_received;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=0;

       if (sigaction(SIGSEGV, &sa, NULL))  {
          sprintf(message, "Unable to setup signal handler for SIGSEGV - This must be fixed - exiting.");
          reportandquit();
       }

/*    plus a handler for SIGPIPE errors    */

   sa.sa_handler = sigpipe_received;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags=0;

       if (sigaction(SIGPIPE, &sa, NULL))  {
          sprintf(message, "Unable to setup signal handler for SIGPIPE - This must be fixed - exiting.");
          reportandquit();
       }

   printf("Done\n");

}
