
#include <time.h>
#include <signal.h>

#define CLOCK_RAVEN 1
#define MSEC 307

timer_t microsleeptimerid;
int microsleepcreated = 1;

void microsleepevent() {}

void microsleepcreate () {
    struct sigevent sigev;
    struct sigaction sa;

    sa.sa_handler = microsleepevent;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags=0;
    if (sigaction(SIGALRM, &sa, NULL)) {
        perror("Unable to setup signal handler for SIGALRM");
        exit(1);
    }

    sigev.sigev_notify = SIGEV_SIGNAL;
    sigev.sigev_signo = SIGALRM;
    if (timer_create(CLOCK_RAVEN, &sigev, &microsleeptimerid)) {
        perror ("Unable to create POSIX.4 timer");
        exit(1);
    }

    microsleepcreated = 1;
}

void microsleepdelete () {

    if (timer_delete(microsleeptimerid)) {
        perror ("Unable to delete POSIX.4 timer");
        exit(1);
    }

    microsleepcreated = 0;
}
        
void microsleep(int n) {
    struct itimerspec timespec;
    sigset_t newsignalSet;
    sigset_t oldsignalSet;

    if (!microsleepcreated) microsleepcreate();

    sigemptyset(&newsignalSet);
    sigaddset(&newsignalSet, SIGALRM);
    sigprocmask(SIG_BLOCK, &newsignalSet, &oldsignalSet);

    timespec.it_value.tv_sec = 0;
    timespec.it_value.tv_nsec = n * MSEC;

    timespec.it_interval.tv_sec =0;
    timespec.it_interval.tv_nsec =0;

    if (timer_settime(microsleeptimerid, 0, &timespec, NULL)) {
        perror("Unable to arm timer");
        exit(1);
    }

    sigemptyset(&newsignalSet);
    (void) sigsuspend(&newsignalSet);

    sigprocmask(SIG_SETMASK, &oldsignalSet, NULL);

    timespec.it_value.tv_sec = 0;
    timespec.it_value.tv_nsec = 0;

    timespec.it_interval.tv_sec =0;
    timespec.it_interval.tv_nsec =0;

    if (timer_settime(microsleeptimerid, 0, &timespec, NULL)) {
        perror("Unable to disarm timer");
        exit(1);
    }
}

void millisleep(int n) {microsleep(n*1000);}
